from PyQt5.QtWidgets import QLineEdit, QHBoxLayout, QLabel, QFileDialog, QPushButton, QWidget, QCheckBox, QVBoxLayout, \
    QDoubleSpinBox, QSpinBox

from album.gui.widgets.argument_widget import ArgumentWidget


def create_gui_string(argument):
    return StringWidget(argument)


def create_gui_file(argument):
    return FileWidget(argument)


def create_gui_directory(argument):
    return FileWidget(argument, QFileDialog.DirectoryOnly)


def create_gui_boolean(argument):
    return BooleanWidget(argument)


def create_gui_float(argument):
    return DoubleWidget(argument)


def create_gui_double(argument):
    return DoubleWidget(argument)


def create_gui_integer(argument):
    return IntegerWidget(argument)


class AbstractWidget(ArgumentWidget):
    def __init__(self, arg):
        super().__init__()
        box = QVBoxLayout()
        self.setLayout(box)
        self.unset_btn = None
        default_value = None
        if "default" in arg:
            default_value = arg["default"]
        arg_name = arg["name"]
        optional = True
        if ("required" not in arg or arg["required"]) and "default" not in arg:
            arg_name += "*"
            optional = False
        self.valuebox = self.create_wrapped_value_widget(default_value, optional)
        nameLb = QLabel("<b>%s</b> %s" % (arg_name, arg["description"]))
        nameLb.setWordWrap(True)
        nameLb.setBuddy(self.valuebox)
        box.addWidget(nameLb)
        box.addWidget(self.valuebox)

    def get_wrapped_value(self):
        if self.unset_btn and self.unset_btn.isChecked():
            return None
        else:
            return self.get_value()

    def get_value(self):
        raise NotImplementedError

    def set_value(self, value):
        raise NotImplementedError

    def create_value_widget(self):
        raise NotImplementedError

    def create_wrapped_value_widget(self, default_value, optional):
        value_widget = self.create_value_widget()
        if optional:
            widget = self._create_optional_widget(value_widget, default_value)
            if default_value:
                self.set_value(default_value)
            return widget
        return value_widget

    def _create_optional_widget(self, widget, default_value, stretch_widget=True):
        res = QWidget()
        layout = QHBoxLayout(res)
        layout.setContentsMargins(0, 0, 0, 0)
        if default_value:
            set_to_default_btn = QPushButton("set to default")
            set_to_default_btn.clicked.connect(lambda: self.set_value(default_value))
            layout.addWidget(set_to_default_btn)
        else:
            self.unset_btn = QCheckBox()
            self.unset_btn.toggled.connect(lambda: widget.setEnabled(not self.unset_btn.isChecked()))
            self.unset_btn.setChecked(True)
            layout.addWidget(self.unset_btn)
            layout.addWidget(QLabel("unset"))
        if stretch_widget:
            layout.addWidget(widget, 1)
        else:
            layout.addStretch()
            layout.addWidget(widget)
        return res


class StringWidget(AbstractWidget):
    def __init__(self, arg):
        super().__init__(arg)

    def create_value_widget(self):
        self.text_field = QLineEdit()
        return self.text_field

    def get_value(self):
        return self.text_field.text()

    def set_value(self, value):
        return self.text_field.setText(value)


class BooleanWidget(AbstractWidget):
    def __init__(self, arg):
        self.checkbox = None
        super().__init__(arg)

    def create_value_widget(self):
        self.checkbox = QCheckBox()
        self.checkbox.setStyleSheet("QCheckBox::indicator{height: 20px; width: 20px;}")
        widget = QWidget()
        layout = QHBoxLayout()
        widget.setLayout(layout)
        layout.addStretch()
        layout.addWidget(self.checkbox)
        widget.layout().setContentsMargins(0, 0, 0, 0)
        return widget

    def get_value(self):
        if self.unset_btn and self.unset_btn.isChecked():
            return None
        return self.checkbox.isChecked()

    def set_value(self, value):
        if value == '0' or value == 'False' or value == 'false':
            value = False
        self.checkbox.setChecked(bool(value))


class DoubleWidget(AbstractWidget):
    def __init__(self, arg):
        super().__init__(arg)

    def create_value_widget(self):
        self.spin_box = QDoubleSpinBox()
        self.spin_box.setDecimals(6)
        self.spin_box.setMaximum(999999)
        return self.spin_box

    def get_value(self):
        return float(self.spin_box.value())

    def set_value(self, value):
        self.spin_box.setValue(float(value))


class IntegerWidget(AbstractWidget):
    def __init__(self, arg):
        super().__init__(arg)

    def create_value_widget(self):
        self.spin_box = QSpinBox()
        self.spin_box.setMaximum(1000000)
        return self.spin_box

    def set_value(self, value):
        self.spin_box.setValue(int(value))

    def get_value(self):
        return int(self.spin_box.value())


class FileWidget(AbstractWidget):
    def __init__(self, arg, file_mode=QFileDialog.AnyFile):
        super().__init__(arg)
        self.file_mode = file_mode

    def create_value_widget(self):
        layout = QHBoxLayout()
        self.btn = QPushButton("Browse")
        self.btn.clicked.connect(self._get_files)
        self.text = QLineEdit()
        layout.addWidget(self.text)
        layout.addWidget(self.btn)
        layout.setContentsMargins(0, 0, 0, 0)
        widget = QWidget()
        widget.setLayout(layout)
        return widget

    def set_value(self):
        #TODO
        pass

    def get_value(self):
        return self.text.text()

    def _get_files(self):
        dlg = QFileDialog()
        dlg.setFileMode(self.file_mode)

        if dlg.exec_():
            self.files = dlg.selectedFiles()
            if self.files:
                self.text.setText(self.files[0])
        self.btn.clearFocus()



from PyQt5.QtCore import pyqtSignal, QThreadPool, QObject
from PyQt5.QtWidgets import QApplication, QDialog, QVBoxLayout, QLabel, QPushButton, \
    QProgressBar, QLineEdit, QCheckBox, QHBoxLayout
from album.api import Album

from album.gui.collection_window import CollectionWindow
from album.gui.dialog.about_dialog import AboutDialog
from album.gui.dialog.solutions_action_dialog import SolutionsActionDialog
from album.gui.dialog.uninstall_solutions_from_catalog_dialog import UninstallSolutionsFromCatalogDialog
from album.gui.dialog.update_album_dialog import UpdateAlbumDialog
from album.gui.dialog.update_catalog_dialog import UpdateCatalogDialog
from album.gui.dialog.update_solutions_dialog import UpdateSolutionDialog
from album.gui.solution_util import generate_install_all_actions
from album.gui.solution_window import SolutionWidget
from album.gui.widgets.util import display_error, display_confirmation, display_info
from album.gui.worker import Worker


class AlbumGUI(QObject):
    collection_changed = pyqtSignal()

    def __init__(self, album_instance):
        super().__init__()
        self.app = QApplication([])
        self.album_instance: Album = album_instance
        self.threadpool = QThreadPool()
        self.open_windows = []
        self.ongoing_processes = {}

    def launch(self, solution=None):
        if solution:
            win = SolutionWidget()
            try:
                self._install_if_needed(win, solution)
            except LookupError:
                display_error("Cannot find solution %s." % solution)
                return
        else:
            win = CollectionWindow(self.album_instance)
            win.show_solution.connect(lambda sol: self.launch_solution(sol))
            win.install_solution.connect(lambda sol: self.install_solution(sol))
            win.uninstall_solution.connect(lambda sol: self.uninstall_solution(sol))
            win.uninstall_solutions.connect(lambda sol: self.uninstall_solutions(sol))
            win.remove_catalog.connect(self.remove_catalog)
            win.add_new_catalog.connect(self.add_catalog)
            win.open_about.connect(self.show_about_dialog)
            win.update_album.connect(self.update_album_dialog)
            win.update_solution_dialog.connect(self.choose_update_solutions)
            win.batch_uninstall_dialog.connect(self.choose_batch_uninstall_catalogs)
            win.update_catalog_dialog.connect(self.choose_update_catalogs)
            self.collection_changed.connect(win.update_model)
        win.show()
        self.app.exec()

    def launch_solution(self, solution):
        win = SolutionWidget()
        try:
            self._install_if_needed(win, solution)
        except LookupError:
            display_error("Cannot find solution %s." % solution)
        self.open_windows.append(win)
        win.show()

    def install_solution(self, solution):
        win = SolutionWidget()
        try:
            solution_data = self.get_database_entry(solution)
            self._install(win, solution, solution_data)
        except LookupError:
            display_error("Cannot find solution %s." % solution)
        self.open_windows.append(win)
        win.show()

    def get_database_entry(self, solution):
        return self.get_unloaded_resolve_result(solution).database_entry()

    def get_unloaded_resolve_result(self, solution):
        return self.album_instance._controller.collection_manager()._resolve(solution)

    def uninstall_solution(self, solution):
        win = SolutionWidget()
        try:
            solution_data = self.get_database_entry(solution)
            win.set_show_solution(solution_data)
            if not self.album_instance.is_installed(solution):
                return
            else:
                win.get_uninstall_widget().continue_signal.connect(lambda: win.close())
                win.show_uninstall()
                worker = Worker(self.album_instance.uninstall, {"solution_to_resolve": solution})
                worker.handler.task_finished.connect(lambda: self._widget_finished(win.get_uninstall_widget()))
                self.ongoing_processes[win] = worker
                worker.handler.new_log.connect(lambda records: win.get_uninstall_widget().update_solution_log(records))
                self.threadpool.start(worker)
        except LookupError:
            display_error("Cannot find solution %s." % solution)
        self.open_windows.append(win)
        win.show()

    def uninstall_solutions(self, actions):
        dialog = SolutionsActionDialog(self.album_instance, actions, "Uninstall solution versions", "Would you like to uninstall the following versions of this solution?", "Uninstall")
        dialog.upgradeDone.connect(self.collection_changed)
        dialog.exec_()

    def _widget_finished(self, widget):
        widget.set_solution_finished()
        self.collection_changed.emit()

    def remove_catalog(self, catalog):
        if display_confirmation("Do you really want to remove catalog %s (%s)?" % (catalog["name"], catalog["src"])):
            try:
                worker = Worker(self.album_instance.remove_catalog_by_name, {"catalog_src": catalog["name"]})
                worker.handler.task_finished.connect(self.collection_changed)
                worker.handler.task_finished.connect(self.catalog_removed)
                worker.handler.new_log.connect(lambda record: self._check_log_for_error(record))
                self.threadpool.start(worker)
            except RuntimeError as e:
                display_error(str(e))

    def add_catalog(self, parent):
        dialog = QDialog(parent)
        layout = QVBoxLayout(dialog)
        input_label = QLabel('Enter the catalog path or URL:')
        line_edit = QLineEdit()
        checkbox = QCheckBox('Install all solutions from catalog')

        ok_button = QPushButton('OK')
        ok_button.clicked.connect(dialog.accept)

        cancel_button = QPushButton('Cancel')
        cancel_button.clicked.connect(dialog.reject)
        btn_layout = QHBoxLayout()
        btn_layout.addWidget(ok_button)
        btn_layout.addWidget(cancel_button)
        layout.addWidget(input_label)
        layout.addWidget(line_edit)
        layout.addWidget(checkbox)
        layout.addLayout(btn_layout)
        if dialog.exec_() == QDialog.Accepted:
            progress_dialog = QDialog(parent)
            layout = QVBoxLayout()
            close_button = QPushButton("Close")
            close_button.setVisible(False)
            close_button.clicked.connect(progress_dialog.close)
            catalog_src = line_edit.text().strip()
            progress_label = QLabel("Adding catalog %s.." % catalog_src)
            layout.addWidget(progress_label)
            progressbar = QProgressBar()
            progressbar.setRange(0, 0)
            layout.addWidget(progressbar)
            layout.addWidget(close_button)
            progress_dialog.setLayout(layout)
            try:

                # Create worker
                worker = Worker(self.album_instance.add_catalog, {"catalog_src": catalog_src})

                worker.handler.task_finished.connect(self.collection_changed)
                worker.handler.task_finished.connect(lambda: progressbar.setVisible(False))
                worker.handler.task_finished.connect(lambda: close_button.setVisible(True))
                if checkbox.isChecked():
                    worker.handler.task_finished.connect(lambda: progress_dialog.close())
                    worker.handler.task_finished.connect(lambda: self.install_all_solutions_from_catalog(catalog_src))
                else:
                    worker.handler.task_finished.connect(lambda: progress_label.setText("Successfully added catalog %s." % catalog_src))

                # Start the worker
                self.threadpool.start(worker)
                progress_dialog.exec_()

            except Exception as e:
                self.album_instance.remove_catalog_by_src(catalog_src)
                display_error(str(e))
                progress_dialog.close()  # Close the progress dialog in case of error

    def install_all_solutions_from_catalog(self, catalog_src):
        install_actions = generate_install_all_actions(self.album_instance, catalog_src)
        dialog = SolutionsActionDialog(self.album_instance, install_actions, "Install all solutions from catalog",
                                       "Successfully added catalog %s. Would you like to install the following solutions from the catalog?" % catalog_src,
                                       "Install")
        dialog.upgradeDone.connect(self.collection_changed)
        dialog.exec_()

    def show_about_dialog(self, parent):
        dialog = AboutDialog(parent)
        dialog.exec_()

    def update_album_dialog(self, parent):
        dialog = UpdateAlbumDialog(parent, self.album_instance.configuration().base_cache_path())
        dialog.exec_()

    def choose_update_catalogs(self, options):
        dialog = UpdateCatalogDialog(options, self.album_instance)
        dialog.upgradeDone.connect(self.collection_changed)
        dialog.exec_()

    def choose_batch_uninstall_catalogs(self, options):
        dialog = UninstallSolutionsFromCatalogDialog(options, self.album_instance)
        dialog.upgradeDone.connect(self.collection_changed)
        dialog.exec_()

    def choose_update_solutions(self):
        dialog = UpdateSolutionDialog(self.album_instance)
        dialog.upgradeDone.connect(self.collection_changed)
        dialog.exec_()

    def catalog_removed(self):
        display_info("Action complete", "Successfully removed catalog.")

    def _install_if_needed(self, win, solution):
        solution_data = self.get_database_entry(solution)
        win.set_show_solution(solution_data)
        if not self.is_installed(solution):
            win.get_pre_install_widget().install_solution.connect(lambda: self._install(win, solution, solution_data))
            win.show_pre_install()
        else:
            self._show_pre_run(win, solution, solution_data)

    def is_installed(self, solution):
        resolve_result = self.get_unloaded_resolve_result(solution)
        return self.album_instance._controller.solutions().is_installed(
            resolve_result.catalog(), resolve_result.coordinates()
        )

    def _install(self, win, solution, solution_data):
        win.set_show_solution(solution_data)
        win.get_install_widget().continue_signal.connect(lambda: self._show_pre_run(win, solution, solution_data))
        win.show_install()

        worker = Worker(self.album_instance.install, {"solution_to_resolve": solution})
        worker.handler.task_finished.connect(lambda: self._widget_finished(win.get_install_widget()))
        self.ongoing_processes[win] = worker
        worker.handler.new_log.connect(lambda records: win.get_install_widget().update_solution_log(records))
        self.threadpool.start(worker)

    def _cancel_installation(self, win, solution, solution_data):
        self.threadpool.cancel(self.ongoing_processes[win])
        self.ongoing_processes.pop(win)
        win.close()

    def _show_pre_run(self, win, solution, solution_data):
        win.get_pre_run_widget().run_solution.connect(lambda: self._run(win, solution))
        win.get_pre_run_widget().cancel_solution.connect(lambda: win.close())
        win.show_pre_run()
        if "args" in solution_data.setup():
            args = solution_data.setup()["args"]
            win.get_pre_run_widget().add_arguments(args)

    def _setup_album(self, album_instance):
        self.album_instance = album_instance

    def dispose(self):
        self.album_instance = None

    def _cancel(self):
        self.dispose()

    def _run(self, win, solution):
        if win.get_pre_run_widget().check_required_fields():
            if self.album_instance:
                values = win.get_pre_run_widget().get_values()
                win.show_run()
                win.get_run_widget().continue_signal.connect(lambda: self.restart_run(win))
                worker = Worker(self.album_instance.run, {"solution_to_resolve": solution,
                                                          "argv": self._construct_args(solution, values)})
                worker.handler.task_finished.connect(lambda: win.get_run_widget().set_solution_finished())
                self.ongoing_processes[win] = worker
                worker.handler.new_log.connect(lambda records: win.get_run_widget().update_solution_log(records))
                self.threadpool.start(worker)

    def restart_run(self, win):
        win.get_run_widget().run_output.clear()
        win.get_run_widget().set_solution_unfinished()
        return win.show_pre_run()

    @staticmethod
    def _build_args(args):
        res = []
        for arg in args:
            res.append("--%s" % arg)
            res.append(args[arg])
        return res

    @staticmethod
    def _construct_args(solution, values):
        res = [solution]
        for name in values:
            res.append("--%s" % name)
            res.append(values[name])
        return res

    @staticmethod
    def _check_log_for_error(record):
        if record.levelname == "ERROR":
            display_error(str(record.msg))


if __name__ == '__main__':
    album = Album.Builder().build()
    album.load_or_create_collection()
    gui = AlbumGUI(album)
    gui.launch()
    # gui.launch(solution="template-python")



import pathlib
from importlib import metadata
from importlib.metadata import PackageNotFoundError

from PyQt5 import QtGui
from PyQt5.QtCore import Qt, pyqtSignal, QSortFilterProxyModel, QEvent
from PyQt5.QtGui import QKeyEvent, QPixmap, QPainter, QFont, QIcon, QColor, QTextCursor
from PyQt5.QtWidgets import QVBoxLayout, QLabel, QLineEdit, QHBoxLayout, QWidget, QMainWindow, QPushButton, QMenu, \
    QAction, QActionGroup, QSplitter, QComboBox, QStyle, QMessageBox, QTextEdit
from album.api import Album
from packaging.version import Version

import album
from album.gui.actions_list_model import ActionsListModel
from album.gui.solution_util import installed, group_solutions_by_version, solution_version, solution_coordinates, \
    full_coordinates, get_newest_prefer_installed
from album.gui.widgets.solution_view import SolutionView
from album.gui.widgets.util import create_btn


class CollectionWindow(QMainWindow):
    # events
    return_pressed = pyqtSignal(QKeyEvent)
    esc_pressed = pyqtSignal(QKeyEvent)
    show_solution = pyqtSignal(str)
    uninstall_solution = pyqtSignal(str)
    uninstall_solutions = pyqtSignal(list)
    install_solution = pyqtSignal(str)
    remove_catalog = pyqtSignal(dict)
    add_new_catalog = pyqtSignal(QWidget)
    open_about = pyqtSignal(QWidget)
    update_album = pyqtSignal(QWidget)
    update_catalog_dialog = pyqtSignal(list)
    update_solution_dialog = pyqtSignal()
    batch_uninstall_dialog = pyqtSignal(list)
    collection_changed = pyqtSignal()

    update_running = pyqtSignal(bool)
    upgrade_running = pyqtSignal(bool)
    available_updates_changed = pyqtSignal(bool)

    class Action:
        def __init__(self, signal, args=None, versions=None, catalog=None):
            self.signal = signal
            self.args = args
            self.versions = versions
            self.catalog = catalog

    def __init__(self, album_instance: Album, parent=None):
        super(CollectionWindow, self).__init__(parent)
        self.selected_solution = None
        self.selected_solution_version = None
        self.album_instance: Album = album_instance
        self.all_catalogs = []
        self.deletable_catalogs = []
        self.list_widget = None
        self.setWindowTitle("Album")
        try:
            self.album_api_version = metadata.version("album-runner")
        except PackageNotFoundError:
            self.album_api_version = metadata.version("album-solution-api")
        self.album_core_version = album.core.__version__
        self.album_gui_version = metadata.version("album-gui")
        self.installed_icon = self.create_icon_from_unicode(chr(0x2713), "#19c18f", size=12)
        self.notinstalled_icon = self.create_icon_from_unicode(chr(0x2717), "#ff6c87", size=12)

        self.search_model = ActionsListModel()
        # self.search_proxy_model = ActionFilterProxyModel()
        self.update_model()

        widget = QWidget(self)
        self.setMinimumWidth(800)
        self.setMinimumHeight(400)
        self.setWindowIcon(QtGui.QIcon(str(pathlib.Path(__file__).parent / 'resources' / 'album-xs.png')))
        self.layout = QVBoxLayout(widget)

        self.layout.setContentsMargins(0, 10, 0, 0)
        self.layout.setSpacing(0)

        self._create_list_widget()
        self._create_search_line()

        top_layout = QSplitter()
        top_layout.addWidget(self._create_center())
        top_layout.addWidget(self._create_right())
        self.layout.addWidget(self.search_line)
        self.layout.addWidget(top_layout, 1)
        self.layout.addLayout(self._create_status_layout())

        self.setCentralWidget(widget)

    def _create_center(self):
        res = QWidget()
        layout = QVBoxLayout(res)
        res.setAutoFillBackground(True)
        # res.setPalette(self.pal)
        res.setContentsMargins(2, 0, 2, 0)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        layout.addWidget(self.list_widget)
        return res

    def _create_search_line(self):
        self.search_line = QLineEdit(self)
        self.search_line.setStyleSheet(
            "QLineEdit { border-radius: 0px; border-top: 1px solid #bbbbbb; border-bottom: 1px solid #bbbbbb;padding: 3px;}")
        self.search_line.installEventFilter(self)
        self.proxy.setFilterCaseSensitivity(False)
        self.search_line.textChanged.connect(self.proxy.setFilterFixedString)
        self.proxy.rowsInserted.connect(self.list_widget.select_first_item)
        self.proxy.rowsRemoved.connect(self.list_widget.select_first_item)
        self.proxy.rowsMoved.connect(self.list_widget.select_first_item)

    def _create_list_widget(self):
        self.proxy = QSortFilterProxyModel(self)
        self.proxy.setSourceModel(self.search_model)
        self.list_widget = SolutionView(self)
        self.list_widget.selectionModel().selectionChanged.connect(self._update_focus_solution)

    def _update_focus_solution(self):
        item = self._get_currently_selected_item()
        if item:
            versions = item.action.versions
            self.version_combo.clear()

            for idx, version in enumerate(versions):
                self.version_combo.addItem(solution_version(version))
                if installed(version):
                    self.version_combo.setItemIcon(idx, self.installed_icon)
                else:
                    self.version_combo.setItemIcon(idx, self.notinstalled_icon)
            # self.version_combo.addItems(solution_version(version) for version in versions)
            # self.version_combo.setEnabled(len(versions) > 1)
            self.selected_solution = get_newest_prefer_installed(versions)
            self.selected_solution_version = solution_version(self.selected_solution)
            self.version_combo.setCurrentText(self.selected_solution_version)
            self._update_focus_solution_version()

    @staticmethod
    def create_icon_from_unicode(unicode_char, color='black', font_name='Arial', size=16):
        pixmap = QPixmap(size, size)
        pixmap.fill(Qt.transparent)
        painter = QPainter(pixmap)
        painter.setPen(QColor(color))
        painter.setFont(QFont(font_name, size))
        painter.drawText(0, 0, size, size, Qt.AlignCenter, unicode_char)
        painter.end()
        return QIcon(pixmap)

    def _update_focus_solution_version(self):
        if self.selected_solution:
            if self.version_combo.currentText():
                self.selected_solution_version = self.version_combo.currentText()

                item = self._get_currently_selected_item()
                versions = item.action.versions
                version_match = [version for version in versions if solution_version(version) == self.selected_solution_version]
                self.selected_solution = version_match[0]
                item.action.args = full_coordinates(self.selected_solution, item.action.catalog)

            self.selected_solution_group.setText(self.selected_solution["setup"]["group"])
            self.selected_solution_name.setText(self.selected_solution["setup"]["name"])

            self.solution_api_warning.setVisible(Version(self.album_api_version) < Version(self.selected_solution["setup"]["album_api_version"]))

            self.selected_solution_description.setPlainText("")
            if "title" in self.selected_solution["setup"] and self.selected_solution["setup"]["title"]:
                self.selected_solution_description.append("~ <b>%s</b>" % self.selected_solution["setup"]["title"])

            if "description" in self.selected_solution["setup"] and self.selected_solution["setup"]["description"]:
                self.selected_solution_description.append(self.selected_solution["setup"]["description"])

            if "cite" in self.selected_solution["setup"] and self.selected_solution["setup"]["cite"]:
                citations = ""
                for citation in self.selected_solution["setup"]["cite"]:
                    citations += citation["text"] + "\n"
                    if "url" in citation:
                        citations += citation["url"] + "\n"
                self.selected_solution_description.append("\nCredits: %s" % citations if citations else "")

            if "solution_creators" in self.selected_solution["setup"]:
                creators = ""
                for creator in self.selected_solution["setup"]["solution_creators"]:
                    creators += creator + "\n"
                self.selected_solution_description.append("\nSolution file written by %s" % creators if creators else "")

            cursor = self.selected_solution_description.textCursor()
            cursor.movePosition(QTextCursor.Start)
            self.selected_solution_description.setTextCursor(cursor)

            self.run_btn.setVisible(True)
        else:
            self.selected_solution_description.setPlainText("")
            self.selected_solution_group.setText("")
            self.selected_solution_name.setText("")
            self.run_btn.setVisible(False)

    def _create_status_layout(self):
        status_layout = QHBoxLayout()
        catalog_status = self._create_status_with_action(
            "catalogs: %s" % len(self.album_instance.get_catalogs_as_dict()["catalogs"]), self._get_catalog_actions())

        solutions_status = self._create_status(
            "solution versions total: %s" % self._get_solution_count(self.album_instance.get_index_as_dict()))
        installed_status = self._create_status_with_action(
            "installed solution versions: %s" % self._get_installed_solution_count(self.album_instance.get_index_as_dict()), self._get_update_solution_actions())
        # running_status = self._create_status("running: 0")
        # updates_status = self._create_status("updates: 0")
        self.collection_changed.connect(lambda: catalog_status.setText(
            "catalogs: %s" % len(self.album_instance.get_catalogs_as_dict()["catalogs"])))
        self.collection_changed.connect(lambda: self._set_actions(catalog_status, self._get_catalog_actions()))
        self.collection_changed.connect(lambda: solutions_status.setText(
            "solutions: %s" % self._get_solution_count(self.album_instance.get_index_as_dict())))
        self.collection_changed.connect(lambda: installed_status.setText(
            "installed: %s" % self._get_installed_solution_count(self.album_instance.get_index_as_dict())))
        # update_btn = self._create_status_with_action("update", self._get_update_actions())

        status_layout.addWidget(catalog_status)
        status_layout.addWidget(solutions_status)
        status_layout.addWidget(installed_status)
        status_layout.addStretch()

        album_label = QPushButton("Album v%s" % (self.album_core_version))
        album_pixmap = QIcon(str(pathlib.Path(__file__).parent / 'resources' / 'album-xs.png'))
        album_label.setIcon(album_pixmap)
        self._set_actions(album_label, self._get_installation_actions())

        status_layout.addWidget(album_label)
        status_layout.setContentsMargins(3, 3, 3, 3)
        status_layout.setSpacing(10)
        return status_layout

    def _get_catalog_actions(self):
        actions = QActionGroup(self)
        add_catalog = QAction("Add catalog")
        add_catalog.triggered.connect(lambda: self.add_new_catalog.emit(self))
        actions.addAction(add_catalog)
        update_catalog = QAction("Update catalog(s)")
        update_catalog.triggered.connect(lambda: self.update_catalog_dialog.emit([catalog["name"] for catalog in self.all_catalogs]))
        actions.addAction(update_catalog)
        for catalog in self.deletable_catalogs:
            add_catalog = QAction("Remove catalog %s" % catalog["name"])
            add_catalog.triggered.connect(self.make_remove_catalog_action(catalog))
            actions.addAction(add_catalog)
        return actions.actions()

    def _get_installation_actions(self):
        actions = QActionGroup(self)
        about = QAction("About")
        about.triggered.connect(lambda: self.open_about.emit(self))
        actions.addAction(about)
        update_album = QAction("Update or reinstall Album (experimental)")
        update_album.triggered.connect(lambda: self.update_album.emit(self))
        actions.addAction(update_album)
        return actions.actions()

    def _get_update_solution_actions(self):
        actions = QActionGroup(self)
        batch_uninstall = QAction("Uninstall all solutions from catalog(s)")
        batch_uninstall.triggered.connect(lambda: self.batch_uninstall_dialog.emit([catalog["name"] for catalog in self.all_catalogs]))
        update_solutions = QAction("Update solution(s)")
        update_solutions.triggered.connect(self.update_solution_dialog)
        actions.addAction(update_solutions)
        actions.addAction(batch_uninstall)
        return actions.actions()

    def make_remove_catalog_action(self, catalog):
        def remove_catalog():
            self.remove_catalog.emit(catalog)
        return remove_catalog

    def _create_action_widget(self):
        action_widget = QWidget()
        action_layout = QHBoxLayout(action_widget)
        action_layout.setContentsMargins(0, 0, 0, 0)
        self.run_btn = create_btn(self, self.run_action, "Run", "Enter")
        self.run_btn.setMinimumWidth(100)
        self.run_btn.setAutoDefault(True)
        self.run_btn.setDefault(True)
        self.search_line.returnPressed.connect(self.run_btn.toggle)
        action_layout.addWidget(self.run_btn)
        self.return_pressed.connect(self.run_btn.keyPressEvent)
        self.esc_pressed.connect(self.close)
        action_widget.setContentsMargins(0, 2, 0, 0)
        return action_widget

    def _create_right(self):
        res_w = QWidget()
        res_w.setFixedWidth(270)

        self.version_combo = QComboBox()
        self.version_combo.currentIndexChanged.connect(self._update_focus_solution_version)

        res = QVBoxLayout(res_w)
        res.setContentsMargins(10, 10, 10, 0)

        self.selected_solution_group = QLabel("")
        self.selected_solution_name = QLabel("")
        self.solution_api_warning = QPushButton("Warning: Potential incompatibility")
        self.solution_api_warning.setIcon(self.style().standardIcon(QStyle.SP_MessageBoxWarning))
        self.solution_api_warning.clicked.connect(self.show_warning_dialog)

        self.selected_solution_description = self.create_readonly_textedit("No description provided.")
        content_widget = QWidget(res_w)
        content_layout = QVBoxLayout(content_widget)
        content_layout.setContentsMargins(2, 2, 2, 2)
        content_layout.addWidget(self.selected_solution_description)
        content_layout.addStretch()
        combo_layout_version = QHBoxLayout()
        combo_layout_version.addWidget(QLabel("Version: "))
        combo_layout_version.addWidget(self.version_combo)
        layout_group = QHBoxLayout()
        layout_group.addWidget(QLabel("Group: "))
        layout_group.addStretch()
        layout_group.addWidget(self.selected_solution_group)
        layout_name = QHBoxLayout()
        layout_name.addWidget(QLabel("Name: "))
        layout_name.addStretch()
        layout_name.addWidget(self.selected_solution_name)

        self.solution_locations = QVBoxLayout()

        res.addLayout(layout_group)
        res.addLayout(layout_name)
        res.addLayout(combo_layout_version)
        res.addWidget(self.selected_solution_description)
        res.addLayout(self.solution_locations)
        res.addWidget(self.solution_api_warning)
        res.addWidget(self._create_action_widget())
        self._update_focus_solution()
        return res_w

    def create_readonly_textedit(self, text):
        textedit = QTextEdit()
        textedit.setPlainText(text)
        textedit.setReadOnly(True)
        textedit.setTextInteractionFlags(Qt.TextSelectableByMouse)  # Allows text selection but not editing
        textedit.setFocusPolicy(Qt.StrongFocus)
        textedit.setTextInteractionFlags(Qt.TextEditorInteraction)
        textedit.setWordWrapMode(1)  # Word wrap
        return textedit

    @staticmethod
    def show_warning_dialog():
        QMessageBox.warning(None, "Warning", "Your Album installation might be too old to install and run this solution. Please update your installation if you run into issues, for example by running the Album Install Wizard again. This will not uninstall or remove any existing solutions or catalogs.")

    @staticmethod
    def _create_status(title):
        label = QLabel("%s" % title)
        return label

    def _create_status_with_action(self, title, actions):
        label = QPushButton("%s" % title, self)
        self._set_actions(label, actions)
        return label

    def _set_actions(self, button, actions):
        menu = QMenu(button)
        # label.setContextMenuPolicy(Qt.ActionsContextMenu)
        menu.addActions(actions)
        button.setMenu(menu)

    def eventFilter(self, source, event):
        if event.type() == QEvent.KeyPress:
            key = event.key()
            if key == Qt.Key_Return:
                self.return_pressed.emit(event)
            if key == Qt.Key_Escape:
                self.esc_pressed.emit(event)
            if source is self.search_line:
                key = event.key()
                if key == Qt.Key_Left or key == Qt.Key_Right or key == Qt.Key_Up or key == Qt.Key_Down:
                    self.list_widget.keyPressEvent(event)
        return super().eventFilter(source, event)

    def run_action(self):
        item = self._get_currently_selected_item()
        action = item.action
        if action.args:
            action.signal.emit(action.args)
        else:
            action.signal.emit()

    def _get_currently_selected_item(self):
        if len(self.list_widget.selectedIndexes()) > 0:
            source_index = self.proxy.mapToSource(self.list_widget.selectedIndexes()[0])
            item = self.search_model.item(source_index.row())
            return item
        return None

    def update_model(self):
        index = self.album_instance.get_index_as_dict()
        actions_list = []
        self.all_catalogs = []
        self.deletable_catalogs = []
        for catalog in index["catalogs"]:

            solutions = group_solutions_by_version(catalog)

            for solution_key in solutions:
                versions = solutions[solution_key]
                solution = get_newest_prefer_installed(versions)
                item = self._create_action_item(catalog, solution, solution_key, versions)
                actions_list.append(item)
        for catalog in index["catalogs"]:
            self.all_catalogs.append(catalog)
            if catalog["deletable"]:
                self.deletable_catalogs.append(catalog)

        actions_list.sort()
        self.search_model.removeRows(0, self.search_model.rowCount())
        for item in actions_list:
            self.search_model.appendRow(item)
        # self.search_model.set_list(actions_list)
        if self.list_widget:
            self.list_widget.select_first_item()
        self.collection_changed.emit()

    def _create_action_item(self, catalog, solution, solution_key, versions):
        coordinates = solution_coordinates(solution)
        if "title" in solution["setup"]:
            title = solution["setup"]["title"]
            if not title:
                title = coordinates
        else:
            title = coordinates
        item = ActionsListModel.ActionItem()
        item.setEditable(False)
        item.setCheckState(2 if installed(solution) else 0)
        item.setText("[%s] %s" % (catalog["name"], title))
        item.text = "[%s] %s" % (catalog["name"], title)
        item.action = CollectionWindow.Action(self.show_solution, full_coordinates(solution, catalog), versions, catalog)
        item.solution_key = solution_key
        return item

    def set_active(self):
        pass

    def set_not_active(self):
        pass

    @staticmethod
    def _get_solution_count(index_dict):
        count = 0
        for catalog in index_dict["catalogs"]:
            count += len(catalog["solutions"])
        return count

    @staticmethod
    def _get_installed_solution_count(index_dict):
        count = 0
        for catalog in index_dict["catalogs"]:
            count += len([solution for solution in catalog["solutions"] if installed(solution)])
        return count



import logging
import platform
import shutil
import subprocess
import sys
import tempfile
from http.client import InvalidURL
from pathlib import Path
from urllib.error import HTTPError
from urllib.request import urlopen

from PyQt5.QtCore import QCoreApplication, pyqtSignal, QThread
from PyQt5.QtNetwork import QNetworkAccessManager
from PyQt5.QtWidgets import QApplication, QDialog, QVBoxLayout, QLabel, QPushButton, QProgressBar, QHBoxLayout, \
    QLineEdit
from album.core.utils.operations.file_operations import force_remove
from album.runner.album_logging import get_active_logger


class Downloader(QThread):

    setTotalProgress = pyqtSignal(int)
    setCurrentProgress = pyqtSignal(int)
    stop_signal = pyqtSignal()
    succeeded = pyqtSignal()
    failed = False
    failed_msg = ""

    def __init__(self, url, filename):
        super().__init__()
        self._url = url
        self._filename = filename
        self.running = True
        self.stop_signal.connect(self.stop)

    def stop(self):  # Method to set running to False
        self.running = False

    def run(self):
        self.failed = True
        try:
            self.download()
            self.failed = False
            self.succeeded.emit()
        except InvalidURL:
            self.failed_msg = "Invalid URL: %s" % self._url
        except HTTPError:
            self.failed_msg = "Cannot download %s" % self._url
        except Exception as e:
            self.failed_msg = str(e)

    def download(self):
        get_active_logger().info("Downloading %s.." % self._url)
        readBytes = 0
        chunkSize = 1024
        with urlopen(self._url) as r:
            # Tell the window the amount of bytes to be downloaded.
            self.setTotalProgress.emit(int(r.info()["Content-Length"]))
            with open(self._filename, "ab") as f:
                while True:
                    if not self.running:
                        break
                    # Read a piece of the file we are downloading.
                    chunk = r.read(chunkSize)
                    # If the result is `None`, that means data is not
                    # downloaded yet. Just keep waiting.
                    if chunk is None:
                        continue
                    # If the result is an empty `bytes` instance, then
                    # the file is complete.
                    elif chunk == b"":
                        break
                    # Write into the local file the downloaded chunk.
                    f.write(chunk)
                    readBytes += chunkSize
                    # Tell the window how many bytes we have received.
                    self.setCurrentProgress.emit(readBytes)


class UpdateAlbumDialog(QDialog):
    startDownload = pyqtSignal(str, Path)

    def __init__(self, parent, album_base_dir):
        super().__init__(parent)
        logging.debug("UpdateDialog initialized")
        self.album_base_dir = album_base_dir
        layout = QVBoxLayout()
        self.setLayout(layout)
        self.manager = QNetworkAccessManager()

        self.info_label = QLabel("This is an experimental feature. Use at your own risk. You can always rerun the install wizard from the website to update or fix an Album installation.\n\nIn order to update your Album installation, the following steps will be executed:\n\n- the Album Installer Wizard will be downloaded\n- this application (Album) will be closed\n- the Album Installer Wizard will be executed\n- Album will be relaunched\n\nThis should not remove or deinstall any existing Album catalog or solution on your system.\n\nDo you want to run this routine?")
        self.info_label.setWordWrap(True)
        layout.addWidget(self.info_label)

        branch_choice = QHBoxLayout()
        self.branch_choice_label = QLabel("album-package branch (don\'t change this if you are unsure)")
        branch_choice.addWidget(self.branch_choice_label)
        self.package_branch = QLineEdit("main")
        branch_choice.addWidget(self.package_branch)
        layout.addLayout(branch_choice)

        self.progress_bar = QProgressBar(self)
        self.progress_bar.setRange(0, 100)
        layout.addWidget(self.progress_bar)
        self.progress_bar.hide()

        action_bar = QHBoxLayout()

        self.yes_button = QPushButton("Yes")
        self.yes_button.clicked.connect(self.prepare_download)
        action_bar.addWidget(self.yes_button)

        self.no_button = QPushButton("No")
        self.no_button.clicked.connect(self.reject)
        action_bar.addWidget(self.no_button)

        layout.addLayout(action_bar)

        self.startDownload.connect(self.download)

    def prepare_download(self):
        logging.debug("Download preparation")
        os_type = platform.system()
        url = f"https://gitlab.com/album-app/plugins/album-package/-/jobs/artifacts/{self.package_branch.text()}/raw/installer/"
        save_path = ""
        if os_type == "Windows":
            url += "album_installer.exe?job=windows_installer_build"
            save_path = "album_installer.exe"
        elif os_type == "Linux":
            url += "album_installer?job=linux_installer_build"
            save_path = "album_installer"
        elif os_type == "Darwin":  # macOS
            url += "album_installer?job=macos_installer_build"
            save_path = "album_installer"

        self.save_path = Path(self.album_base_dir).joinpath(save_path)
        self.save_path_tmp = Path(self.album_base_dir).joinpath("tmp_" + str(save_path))
        self.startDownload.emit(url, self.save_path_tmp)

    def download(self, url, save_path):
        self.info_label.setText("Downloading the Album Installer Wizard...")
        self.progress_bar.show()
        self.branch_choice_label.setVisible(False)
        self.package_branch.setVisible(False)
        self.yes_button.setEnabled(False)
        self.no_button.setEnabled(True)
        self.no_button.setText("Cancel")
        self.no_button.disconnect()
        self.no_button.clicked.connect(self.cancel_download)

        self.downloader = Downloader(url, save_path)
        # Connect the signals which send information about the download
        # progress with the proper methods of the progress bar.
        self.downloader.setTotalProgress.connect(self.progress_bar.setMaximum)
        self.downloader.setCurrentProgress.connect(self.progress_bar.setValue)
        # Qt will invoke the `succeeded()` method when the file has been
        # downloaded successfully and `downloadFinished()` when the
        # child thread finishes.
        # self.downloader.succeeded.connect(self.downloadSucceeded)
        self.downloader.finished.connect(self.download_finished)
        self.downloader.start()

    def cancel_download(self):
        if hasattr(self, 'downloader'):
            self.downloader.stop_signal.emit()
            self.downloader.wait()
            self.close()

    def download_finished(self):
        self.progress_bar.hide()
        self.no_button.setEnabled(True)
        self.no_button.disconnect()
        self.no_button.clicked.connect(self.reject)
        if self.downloader.failed:
            self.info_label.setText(
                "Download failed. %s" % self.downloader.failed_msg)
            self.no_button.setText("Close")
            self.yes_button.setEnabled(False)
        else:
            self.no_button.setText("No")
            self.info_label.setText(
                "Download complete. Ready to update? This application will close, the installation wizard will run and Album will relaunch.")
            self.yes_button.setText("Update Now")
            self.yes_button.setEnabled(True)
            self.yes_button.clicked.connect(self.update)
        del self.downloader

    def update(self):
        QCoreApplication.quit()
        command = self.save_path
        shutil.copy(self.save_path_tmp, command)
        self.make_executable(command)
        force_remove(self.save_path_tmp)

        os_type = platform.system()
        full_command = f"\"{command}; echo 'Press Enter to continue.'; read\""

        if os_type == "Windows":
            subprocess.Popen(["cmd.exe", "/k", full_command])
        elif os_type == "Darwin":  # macOS
            subprocess.Popen(["osascript", "-e", f'tell app "Terminal" to do script "{full_command}"'])
        else:  # Assuming Linux
            terminals = ["gnome-terminal", "konsole", "xfce4-terminal", "xterm"]
            for terminal in terminals:
                try:
                    subprocess.Popen([terminal, "--", "bash", "-c", command])
                    break
                except FileNotFoundError:
                    continue
        sys.exit(0)

    def make_executable(self, path):
        import os
        mode = os.stat(path).st_mode
        mode |= (mode & 0o444) >> 2  # copy R bits to X
        os.chmod(path, mode)


if __name__ == "__main__":
    logging.debug("Application started")
    app = QApplication([])
    temp_dir = tempfile.mkdtemp()
    dialog = UpdateAlbumDialog(None, temp_dir)
    dialog.exec_()
    # not deleting the tmp dir because the wizard is installed there and executed after the python process is killed...



from album.api import Album

from album.gui.app import AlbumGUI
from album.gui.create_shortcut import create_shortcut


def launch_gui(album_instance: Album, args):
    gui = _create_gui(album_instance)
    solution = args.solution
    if solution:
        gui.launch(solution=solution)
    else:
        gui.launch()
    gui.dispose()


def _create_gui(album_instance):
    return AlbumGUI(album_instance)


def create_parser_gui(parser):
    p = parser.create_command_parser('gui', launch_gui, 'Launch the Album GUI.')
    p.add_argument('solution', type=str,
                   help='path for the solution file or coordinates of the solution (group:name:version)', nargs='?')


def create_parser_shortcut(parser):
    p = parser.create_command_parser('add-shortcut', create_shortcut, 'Create shortcut for Album.')
    p.add_argument('solution', type=str,
                   help='path for the solution file or coordinates of the solution (group:name:version)', nargs='?')



#!/usr/bin/env python
"""
Create desktop shortcuts for Windows
"""
import os

import win32com.client
from win32com.shell import shell, shellcon

from . import UserFolders
from .shortcut import shortcut

scut_ext = "lnk"
ico_ext = ("ico",)

_WSHELL = win32com.client.Dispatch("Wscript.Shell")


# Windows Special Folders
# see: https://docs.microsoft.com/en-us/windows/win32/shell/csidl


def get_homedir():
    """Return home directory:
    note that we return CSIDL_PROFILE, not
    CSIDL_APPDATA, CSIDL_LOCAL_APPDATA,  or CSIDL_COMMON_APPDATA
    """
    return shell.SHGetFolderPath(0, shellcon.CSIDL_PROFILE, None, 0)


def get_desktop():
    """Return user Desktop folder"""
    return shell.SHGetFolderPath(0, shellcon.CSIDL_DESKTOP, None, 0)


def get_startmenu():
    """Return user Start Menu Programs folder
    note that we return CSIDL_PROGRAMS not CSIDL_COMMON_PROGRAMS
    """
    return shell.SHGetFolderPath(0, shellcon.CSIDL_PROGRAMS, None, 0)


def get_folders():
    """get user-specific folders

    Returns:
    -------
    Named tuple with fields 'home', 'desktop', 'startmenu'

    Example:
    -------
    >>> from pyshortcuts import get_folders
    >>> folders = get_folders()
    >>> print("Home, Desktop, StartMenu ",
    ...       folders.home, folders.desktop, folders.startmenu)
    """
    return UserFolders(get_homedir(), get_desktop(), get_startmenu())


def make_shortcut(
    script,
    name=None,
    description=None,
    icon=None,
    folder=None,
    terminal=True,
    desktop=True,
    startmenu=True,
    executable=None,
):
    """create shortcut

    Arguments:
    ---------
    script      (str) path to script, may include command-line arguments
    name        (str, None) name to display for shortcut [name of script]
    description (str, None) longer description of script [`name`]
    icon        (str, None) path to icon file [python icon]
    folder      (str, None) subfolder of Desktop for shortcut [None] (See Note 1)
    terminal    (bool) whether to run in a Terminal [True]
    desktop     (bool) whether to add shortcut to Desktop [True]
    startmenu   (bool) whether to add shortcut to Start Menu [True] (See Note 2)
    executable  (str, None) name of executable to use [this Python] (see Note 3)

    Notes:
    ------
    1. `folder` will place shortcut in a subfolder of Desktop and/or Start Menu
    2. Start Menu does not exist for Darwin / MacOSX
    3. executable defaults to the Python executable used to make shortcut.
    """
    userfolders = get_folders()

    scut = shortcut(
        script,
        userfolders,
        name=name,
        description=description,
        folder=folder,
        icon=icon,
    )

    for create, folder in (
        (desktop, scut.desktop_dir),
        (startmenu, scut.startmenu_dir),
    ):
        if create:
            if not os.path.exists(folder):
                os.makedirs(folder)
            dest = os.path.normpath(os.path.join(folder, scut.target))

            wscript = _WSHELL.CreateShortCut(dest)
            wscript.Targetpath = '"%s"' % executable
            # "/C" means run and terminate. otherwise, the cmd will interpret the script as a parameter, which does not exist, and just launches a cmd
            wscript.Arguments = "/C " + script
            wscript.WorkingDirectory = userfolders.home
            wscript.WindowStyle = 7  # 7 means Minimized window at startup
            wscript.Description = scut.description
            wscript.IconLocation = scut.icon
            wscript.save()

    return scut



import pathlib
from importlib import metadata
from importlib.metadata import PackageNotFoundError

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QPixmap, QIcon
from PyQt5.QtWidgets import QVBoxLayout, QLabel, QDialog, QPushButton, QHBoxLayout

import album


class AboutDialog(QDialog):
    def __init__(self, parent):
        super().__init__(parent)

        try:
            self.album_api_version = metadata.version("album-runner")
        except PackageNotFoundError:
            self.album_api_version = metadata.version("album-solution-api")
        self.album_core_version = album.core.__version__
        self.album_gui_version = metadata.version("album-gui")

        layout = QVBoxLayout()

        self.setWindowTitle("About Album")

        versions = QLabel("Versions: album v%s, album-gui v%s, max. album_api_version %s" % (self.album_core_version, self.album_gui_version, self.album_api_version))

        disclaimer = QLabel("You are using Album at your own risk. Album will run any code included in a solution without being able to check if it will harm your system, please be aware that this code has access to your system and data.")
        disclaimer.setWordWrap(True)

        maintenance = QLabel("Album is invented, maintained and developed by Helmholtz Imaging.")
        maintenance.setWordWrap(True)

        icon_layout = QHBoxLayout()
        icon_layout.setAlignment(Qt.AlignLeft)
        icon_size = 50
        hi_label = QLabel()
        hi_pixmap = QPixmap(str(pathlib.Path(__file__).parent.parent / 'resources' / 'hi.png'))
        hi_label.setPixmap(hi_pixmap.scaled(icon_size, icon_size, Qt.KeepAspectRatio, Qt.SmoothTransformation))
        icon_layout.addWidget(hi_label)

        album_pixmap = QPixmap(str(pathlib.Path(__file__).parent.parent / 'resources' / 'album-xs.png'))
        album_label = QLabel()
        album_label.setPixmap(album_pixmap.scaled(icon_size, icon_size, Qt.KeepAspectRatio, Qt.SmoothTransformation))
        icon_layout.addWidget(album_label)
        
        ok_button = QPushButton("Close")
        ok_button.clicked.connect(self.accept)
        layout.addLayout(icon_layout)
        album_website = QLabel("<a href='https://album.solutions'>https://album.solutions</a><br/><a href='https://helmholtz-imaging.de'>https://helmholtz-imaging.de</a>")
        album_website.setOpenExternalLinks(True)

        layout.addWidget(album_website)
        layout.addWidget(versions)
        layout.addWidget(maintenance)
        layout.addWidget(disclaimer)
        layout.addWidget(ok_button)

        self.setLayout(layout)


from PyQt5.QtCore import QThread, pyqtSignal
from album.api import Album

from album.gui.dialog.update_action_item import UpdateActionItem, UpdateAction
from album.gui.solution_util import full_coordinates
from album.gui.widgets.util import display_error


class SolutionActionProcessThread(QThread):
    processFinished = pyqtSignal(UpdateActionItem)
    processFailed = pyqtSignal(UpdateActionItem)

    def __init__(self, item, album_instance: Album):
        super().__init__()
        self.item = item
        self.album_instance = album_instance

    def run(self):
        try:
            if self.item.update_action == UpdateAction.INSTALL:
                self.album_instance.install(full_coordinates(self.item.solution, self.item.catalog))
            if self.item.update_action == UpdateAction.UNINSTALL:
                self.album_instance.uninstall(full_coordinates(self.item.solution, self.item.catalog))
            self.processFinished.emit(self.item)
        except Exception as e:
            print(f"An error occurred: {e}")
            display_error(f"An error occurred: {e}")
            self.processFailed.emit(self.item)



import platform
import sys
from pathlib import Path

from album.api import Album

from album.gui.include.pyshortcuts import make_shortcut
from album.core.controller.micromamba_manager import get_mamba_exe


def get_icon_path():
    if platform.system() == "Windows":
        return str(Path(__file__).parent / "resources" / "album_icon_windows.ico")
    elif platform.system() == "Darwin":
        return str(Path(__file__).parent / "resources" / "album_icon_macos.icns")
    elif platform.system() == "Linux":
        return str(Path(__file__).parent / "resources" / "album_icon_linux.png")


def create_shortcut(album_instance: Album, args=None, command="album gui"):
    album_base_path = album_instance.configuration().base_cache_path()
    album_environment_path = album_base_path.joinpath("envs", "album")
    if not Path(album_environment_path).exists():
        # if album is installed at a different place, this should catch it
        album_environment_path = sys.prefix
    micromamba_executable = get_mamba_exe(
        album_instance.configuration().micromamba_base_path()
    )
    package_manager = micromamba_executable
    solution_environments_path = album_instance.configuration().environments_path()
    _create_shortcut_from_environment_and_executable(
        album_base_path,
        album_environment_path,
        solution_environments_path,
        package_manager,
        command,
    )


def _create_shortcut_from_environment_and_executable(
    album_base_path,
    album_environment_path,
    environments_location,
    package_manager,
    command,
):
    env_vars = {
        "ALBUM_BASE_CACHE_PATH": str(album_base_path),
    }
    if "micromamba" in str(package_manager).lower():
        env_vars["MAMBA_ROOT_PREFIX"] = str(environments_location)

    icon_path = get_icon_path()

    if platform.system() == "Windows":
        script_command = (
            f'{package_manager} run -p "{album_environment_path}" {command}'
        )
        script_file_path = Path(album_base_path).joinpath("Album.bat")
        with open(script_file_path, "w") as f:
            for var, value in env_vars.items():
                f.write(f"set {var}={value}\n")
            f.write(script_command + "\n")
        make_shortcut(
            str(script_file_path),
            name="Album",
            icon=icon_path,
            terminal=False,
            executable="C:\\Windows\\System32\\cmd.exe",
        )
    else:  # For MacOS and Linux, keep as it is
        script_command = f"run -p {album_environment_path} {command}"
        envs = ["env"]
        envs += [f"{var}={value}" for var, value in env_vars.items()]
        exec = " ".join(envs + [str(package_manager)])
        if platform.system() == "Darwin":
            exec = "'%s'" % exec
            script_command = "'%s'" % script_command
        make_shortcut(
            script_command,
            name="Album",
            icon=icon_path,
            terminal=False,
            executable=exec,
        )


def write_batch_file(batch_file_path, env_vars, command):
    with open(batch_file_path, "w") as f:
        for var, value in env_vars.items():
            f.write(f"set {var}={value}\n")
        f.write(command + "\n")



import sys

from PyQt5.QtCore import pyqtSignal, Qt
from PyQt5.QtGui import QColor, QBrush, QPalette
from PyQt5.QtWidgets import (
    QDialog, QApplication, QVBoxLayout, QHBoxLayout,
    QPushButton, QLabel, QTableWidget, QTableWidgetItem, QCheckBox, QWidget
)

from album.gui.dialog.update_action_item import UpdateActionItem, UpdateAction
from album.gui.solution_util import full_coordinates, solution_coordinates


class CheckListTableDialog(QDialog):
    actionSubmitted = pyqtSignal(list)

    def __init__(self, options, title="", question="", action_name="", enable_disable_options=False):
        super().__init__()
        self.setWindowTitle(title)
        self.options = options
        self.initUI(question, action_name, enable_disable_options)

    def initUI(self, question, action_name, enable_disable_options):
        self.tableWidget = QTableWidget()
        self.tableWidget.setRowCount(len(self.options))
        self.tableWidget.setColumnCount(4)
        self.tableWidget.setHorizontalHeaderLabels(["Select", "Action", "Catalog", "Solution"])
        self.tableWidget.horizontalHeader().hide()
        self.tableWidget.verticalHeader().hide()
        self.tableWidget.setColumnWidth(0, 40)
        self.tableWidget.setColumnWidth(1, 100)
        self.tableWidget.horizontalHeader().setStretchLastSection(True)

        self.uninstallCheckbox = QCheckBox("Uninstall old version of solutions")
        self.uninstallCheckbox.setChecked(True)
        self.uninstallCheckbox.stateChanged.connect(self.updateRows)

        self.installCheckbox = QCheckBox("Update installed solutions to newest version")
        self.installCheckbox.setChecked(True)
        self.installCheckbox.stateChanged.connect(self.updateRows)

        for i, option in enumerate(self.options):
            checkBoxWidget = QWidget()
            layout = QHBoxLayout(checkBoxWidget)
            layout.setContentsMargins(3, 0, 0, 0)
            layout.setAlignment(Qt.AlignCenter)

            checkbox = QCheckBox()
            checkbox.setChecked(True)
            layout.addWidget(checkbox)

            self.tableWidget.setCellWidget(i, 0, checkBoxWidget)
            action_item = QTableWidgetItem(option.update_action.name)
            self.tableWidget.setItem(i, 1, action_item)
            solution_item = QTableWidgetItem(solution_coordinates(option.solution))
            self.tableWidget.setItem(i, 3, solution_item)
            catalog_item = QTableWidgetItem(str(option.catalog["name"]))
            self.tableWidget.setItem(i, 2, catalog_item)

            self.colorRow(i, option)

        self.tableWidget.setAutoFillBackground(False)
        palette = self.tableWidget.palette()
        palette.setColor(QPalette.Base, Qt.transparent)
        self.tableWidget.setPalette(palette)
        self.tableWidget.setShowGrid(False)
        self.tableWidget.setAlternatingRowColors(True)
        # self.tableWidget.setStyleSheet("border: 0px;")

        hbox = QHBoxLayout()
        self.actionButton = QPushButton(action_name)
        self.actionButton.clicked.connect(self.applyAction)
        self.cancelButton = QPushButton("Cancel")
        self.cancelButton.clicked.connect(self.reject)
        hbox.addWidget(self.actionButton)
        hbox.addWidget(self.cancelButton)

        custom_layout = QVBoxLayout()
        vbox2 = QVBoxLayout()
        q_label = QLabel(question)
        q_label.setWordWrap(True)
        vbox2.addWidget(q_label)
        if enable_disable_options:
            vbox2.addWidget(self.installCheckbox)
            vbox2.addWidget(self.uninstallCheckbox)
        vbox2.addLayout(custom_layout)
        vbox2.addWidget(self.tableWidget)
        vbox2.addLayout(hbox)

        self.setLayout(vbox2)
        self.updateRows()

    def colorRow(self, i, option):
        action_item = self.tableWidget.item(i, 1)

        if option.update_action == UpdateAction.INSTALL:
            action_item.setForeground(QBrush(QColor('green')))
        elif option.update_action == UpdateAction.UNINSTALL:
            action_item.setForeground(QBrush(QColor('red')))

    def updateRows(self):
        for i in range(self.tableWidget.rowCount()):
            action_item = self.tableWidget.item(i, 1)
            checkbox = self.tableWidget.cellWidget(i, 0)

            enabled = True
            if action_item.text() == "UNINSTALL":
                enabled = self.uninstallCheckbox.isChecked()
            elif action_item.text() == "INSTALL":
                enabled = self.installCheckbox.isChecked()

            checkbox.setEnabled(enabled)
            self.tableWidget.item(i, 1).setFlags(
                self.tableWidget.item(i, 1).flags() | Qt.ItemIsEnabled if enabled else self.tableWidget.item(i, 1).flags() & ~Qt.ItemIsEnabled)
            self.tableWidget.item(i, 2).setFlags(
                self.tableWidget.item(i, 2).flags() | Qt.ItemIsEnabled if enabled else self.tableWidget.item(i, 2).flags() & ~Qt.ItemIsEnabled)
            self.tableWidget.item(i, 3).setFlags(
                self.tableWidget.item(i, 3).flags() | Qt.ItemIsEnabled if enabled else self.tableWidget.item(i, 2).flags() & ~Qt.ItemIsEnabled)

            if enabled:
                self.colorRow(i, self.options[i])
            else:
                action_item.setForeground(QBrush(QColor('lightgrey')))

    def applyAction(self):
        checked_items = self.getCheckedItems()
        self.actionSubmitted.emit(checked_items)
        self.accept()

    def getCheckedItems(self):
        checked_items = []
        for i in range(self.tableWidget.rowCount()):
            checkbox_widget = self.tableWidget.cellWidget(i, 0)
            layout = checkbox_widget.layout()
            checkbox = layout.itemAt(0).widget()

            if checkbox.isChecked() and checkbox.isEnabled():
                checked_items.append(self.options[i])

        return checked_items


if __name__ == "__main__":
    app = QApplication(sys.argv)

    options = [
        UpdateActionItem(UpdateAction.NONE, "Solution 1"),
        UpdateActionItem(UpdateAction.UNINSTALL, "Solution 2"),
        UpdateActionItem(UpdateAction.INSTALL, "Solution 3"),
        UpdateActionItem(UpdateAction.INSTALL, "Solution 3"),
        UpdateActionItem(UpdateAction.INSTALL, "Solution 3"),
        UpdateActionItem(UpdateAction.INSTALL, "Solution 3"),
        UpdateActionItem(UpdateAction.INSTALL, "Solution 3"),
        UpdateActionItem(UpdateAction.INSTALL, "Solution 3"),
        UpdateActionItem(UpdateAction.INSTALL, "Solution 3"),
        UpdateActionItem(UpdateAction.INSTALL, "Solution 3"),
        UpdateActionItem(UpdateAction.INSTALL, "Solution 3"),
        UpdateActionItem(UpdateAction.INSTALL, "Solution 3"),
        UpdateActionItem(UpdateAction.INSTALL, "Solution 3"),
        UpdateActionItem(UpdateAction.INSTALL, "Solution 3"),
        UpdateActionItem(UpdateAction.INSTALL, "Solution 3"),
        UpdateActionItem(UpdateAction.INSTALL, "Solution 3"),
    ]

    dialog = CheckListTableDialog(
        options, title="Check List Dialog", question="Please select options:", action_name="Action"
    )

    dialog.tableWidget.horizontalHeader().setStretchLastSection(True)

    if dialog.exec_() == QDialog.Accepted:
        checked_items = dialog.getCheckedItems()
        for item in checked_items:
            print(f"Action: {item.update_action.name}, Solution: {item.solution}")

    sys.exit(app.exec_())



from PyQt5.QtCore import pyqtSignal, QEvent, Qt
from PyQt5.QtGui import QKeyEvent
from PyQt5.QtWidgets import QWidget, QVBoxLayout, QLabel, QStackedLayout, QMainWindow
from album.core.api.model.collection_index import ICollectionIndex

from album.gui.widgets.pre_install_solution_widget import PreInstallSolutionWidget
from album.gui.widgets.pre_run_solution_widget import PreRunSolutionWidget
from album.gui.widgets.solution_status_widget import SolutionStatus


class SolutionWidget(QMainWindow):
    class UninstallSolutionWidget(SolutionStatus):
        def _get_progress_text(self):
            return "Uninstalling..."

        def _get_continue_text(self):
            return "Close"

    class InstallSolutionWidget(SolutionStatus):
        def _get_progress_text(self):
            return "Installing..."

        def _get_continue_text(self):
            return "Run"

    class RunSolutionWidget(SolutionStatus):
        def _get_progress_text(self):
            return "Running..."

        def _get_continue_text(self):
            return "Restart"

    closing = pyqtSignal()
    return_pressed = pyqtSignal(QKeyEvent)
    esc_pressed = pyqtSignal(QKeyEvent)

    INDEX_PRE_INSTALL_SOLUTION = 0
    INDEX_INSTALL_SOLUTION = 1
    INDEX_PRE_RUN_SOLUTION = 2
    INDEX_RUN_SOLUTION = 3
    INDEX_UNINSTALL_SOLUTION = 4

    def __init__(self):
        super().__init__()
        widget = QWidget()
        self.layout = QVBoxLayout(widget)
        self.setCentralWidget(widget)
        self.solution_header = QWidget(self)
        self.stacked_layout = QStackedLayout()
        self.pre_install_widget = PreInstallSolutionWidget(self)
        self.install_widget = SolutionWidget.InstallSolutionWidget(self)
        self.uninstall_widget = SolutionWidget.UninstallSolutionWidget(self)
        self.pre_run_widget = PreRunSolutionWidget(self)
        self.run_widget = SolutionWidget.RunSolutionWidget(self)
        self.layout.addWidget(self.solution_header)
        self.widgets = [self.pre_install_widget, self.install_widget, self.pre_run_widget, self.run_widget,
                        self.uninstall_widget]
        for widget in self.widgets:
            self.stacked_layout.addWidget(widget)
        self.layout.addLayout(self.stacked_layout)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.stacked_layout.setCurrentIndex(0)
        self.pre_install_widget.cancel_solution.connect(lambda: self.close())
        self.esc_pressed.connect(lambda: self.close())
        self.installEventFilter(self)

    def eventFilter(self, source, event):
        if event.type() == QEvent.KeyPress:
            key = event.key()
            if key == Qt.Key_Return:
                self.return_pressed.emit(event)
            if key == Qt.Key_Escape:
                self.esc_pressed.emit(event)
        return super(SolutionWidget, self).eventFilter(source, event)

    def closeEvent(self, *args, **kwargs):
        super(QMainWindow, self).closeEvent(*args, **kwargs)
        self.closing.emit()

    def set_show_solution(self, solution: ICollectionIndex.ICollectionSolution):
        new_header = QWidget()
        layout = QVBoxLayout()
        new_header.setLayout(layout)
        solution_setup = solution.setup()
        if "title" in solution_setup and solution_setup["title"]:
            title = "%s (version %s)" % (solution_setup["title"], solution_setup["version"])
        else:
            title = "%s:%s:%s" % (solution_setup["group"], solution_setup["name"], solution_setup["version"])
        title_label = QLabel(title)
        title_label.setWordWrap(True)
        title_label.setStyleSheet("font-weight: bold;")
        layout.addWidget(title_label)
        if "description" in solution_setup and solution_setup["description"]:
            description = solution_setup["description"]
        else:
            description = "No description provided."
        description_label = QLabel(description)
        description_label.setStyleSheet("color: rgba(0,0,0,0.6)")
        description_label.setWordWrap(True)
        layout.addWidget(description_label)
        self.layout.replaceWidget(self.solution_header, new_header)
        self.solution_header = new_header

    def show_pre_install(self):
        self._set_active(self.INDEX_PRE_INSTALL_SOLUTION)

    def show_install(self):
        self._set_active(self.INDEX_INSTALL_SOLUTION)

    def show_uninstall(self):
        self._set_active(self.INDEX_UNINSTALL_SOLUTION)

    def show_pre_run(self):
        self._set_active(self.INDEX_PRE_RUN_SOLUTION)

    def show_run(self):
        self._set_active(self.INDEX_RUN_SOLUTION)

    def get_pre_install_widget(self):
        return self.pre_install_widget

    def get_install_widget(self):
        return self.install_widget

    def get_uninstall_widget(self):
        return self.uninstall_widget

    def get_pre_run_widget(self):
        return self.pre_run_widget

    def get_run_widget(self):
        return self.run_widget

    def _set_active(self, index):
        self.stacked_layout.setCurrentIndex(index)
        for i, widget in enumerate(self.widgets):
            if index == i:
                widget.set_active()
            else:
                widget.set_not_active()

    def set_active(self):
        pass

    def set_not_active(self):
        pass



import sys

from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import (
    QDialog, QApplication, QCheckBox, QVBoxLayout, QHBoxLayout,
    QPushButton, QScrollArea, QWidget, QLabel
)


class CheckListDialog(QDialog):
    actionSubmitted = pyqtSignal(list)

    def __init__(self, options, title="", question="", action_name=""):
        super().__init__()
        self.setWindowTitle(title)
        self.options = options
        self.checkboxes = []
        self.availableUpdates = []
        self.custom_layout = self.initUI(question, action_name)

    def initUI(self, question, action_name):
        scrollArea = QScrollArea()
        scrollWidget = QWidget()
        vbox = QVBoxLayout(scrollWidget)
        for option in self.options:
            checkbox = QCheckBox(option)
            vbox.addWidget(checkbox)
            self.checkboxes.append(checkbox)
        scrollArea.setWidget(scrollWidget)
        scrollArea.setWidgetResizable(True)

        hbox = QHBoxLayout()
        selectAllButton = QPushButton("Select All")
        selectAllButton.clicked.connect(self.selectAll)
        deselectAllButton = QPushButton("Deselect All")
        deselectAllButton.clicked.connect(self.deselectAll)
        self.actionButton = QPushButton(action_name)
        self.actionButton.clicked.connect(self.applyAction)
        self.cancelButton = QPushButton("Cancel")
        self.cancelButton.clicked.connect(self.reject)
        hbox.addWidget(selectAllButton)
        hbox.addWidget(deselectAllButton)
        hbox.addWidget(self.actionButton)
        hbox.addWidget(self.cancelButton)
        custom_layout = QVBoxLayout()
        vbox2 = QVBoxLayout()
        vbox2.addWidget(QLabel(question))
        vbox2.addLayout(custom_layout)
        vbox2.addWidget(scrollArea)
        vbox2.addLayout(hbox)

        self.setLayout(vbox2)
        return custom_layout

    def selectAll(self):
        for checkbox in self.checkboxes:
            checkbox.setChecked(True)

    def deselectAll(self):
        for checkbox in self.checkboxes:
            checkbox.setChecked(False)

    def applyAction(self):
        self.actionSubmitted.emit(self.getCheckedItems())

    def getCheckedItems(self):
        checked_items = []
        for checkbox in self.checkboxes:
            if checkbox.isChecked():
                checked_items.append(checkbox.text())
        return checked_items


if __name__ == "__main__":
    app = QApplication(sys.argv)
    options = ["Option " + str(i) for i in range(1, 51)]
    dialog = CheckListDialog(
        options, title="Check List Dialog", question="Please select options:", action_name="Action"
    )
    if dialog.exec_() == QDialog.Accepted:
        checked_items = dialog.getCheckedItems()
        print("Activated choices: ", checked_items)
    sys.exit(app.exec_())


from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import QHBoxLayout, QWidget, QVBoxLayout, QLabel

from album.gui.widgets.util import create_btn


class PreInstallSolutionWidget(QWidget):
    # events
    install_solution = pyqtSignal()
    cancel_solution = pyqtSignal()

    def __init__(self, parent):
        super().__init__(parent)
        self.layout = QVBoxLayout(self)
        question = QLabel("This solution isn't installed. Would you like to install it?")
        question.setWordWrap(True)
        self.layout.addWidget(question, 1)
        self.layout.addWidget(self._create_actions_box())
        self._widgets = {}
        self._required = []

    def _create_actions_box(self):
        res = QWidget(self)
        actions_layout = QHBoxLayout(res)
        actions_layout.setContentsMargins(0, 0, 0, 0)
        actions_layout.addWidget(create_btn(self, self.cancel_solution, "Cancel", "Esc"))
        self.install_btn = create_btn(self, self.install_solution, "Install", "Enter")
        actions_layout.addWidget(self.install_btn)
        return res

    def set_active(self):
        self.install_btn.setDefault(True)
        self.install_btn.setAutoDefault(True)
        self.install_btn.setFocus()

    def set_not_active(self):
        self.install_btn.setDefault(False)
        self.install_btn.setAutoDefault(False)



import logging
import platform
import random
import traceback

from PyQt5.QtCore import QRunnable, pyqtSignal, QObject
from album.core.api.model.task import ILogHandler
from album.core.model.task import Task
from album.runner import album_logging
from album.runner.album_logging import get_active_logger, LogLevel


class Worker(QRunnable):

    class LogHandler(ILogHandler, QObject):
        new_log = pyqtSignal(logging.LogRecord)
        task_finished = pyqtSignal()
        task_failed = pyqtSignal()

        def __init__(self):
            QObject.__init__(self)
            ILogHandler.__init__(self)
            self._records = []

        def emit(self, record: logging.LogRecord) -> None:
            self._records.append(record)
            self.new_log.emit(record)

        def records(self):
            return self._records

    def __init__(self, method, method_args):
        super().__init__()
        self.task = Task(method, method_args)
        self.task.set_id(random.randint(0, 10000))
        self.handler = Worker.LogHandler()

    def run(self):
        get_active_logger().debug(f"starting task {self.task.id()}...")
        logger = album_logging.configure_logging("task" + str(self.task.id()), loglevel=LogLevel.INFO)
        logger.addHandler(self.handler)
        self.task.set_log_handler(self.handler)
        try:
            if platform.system() == 'Windows':
                import pythoncom
                pythoncom.CoInitialize()
            self.task.method()(**self.task.args())
            get_active_logger().debug(f"finished task {self.task.id()}.")
            self.handler.task_finished.emit()
        except Exception as e:
            traceback.print_exc()
            logger.error(str(e))
            # TODO set window to failed
        finally:
            if platform.system() == 'Windows':
                pythoncom.CoUninitialize()
            logger.removeHandler(self.handler)
            album_logging.pop_active_logger()



from enum import Enum


class UpdateAction(Enum):
    NONE = 0,
    UNINSTALL = 1,
    INSTALL = 2


class UpdateActionItem:

    def __init__(self, update_action: UpdateAction = UpdateAction.NONE, catalog = None, solution=None, iteration=0):
        self.update_action = update_action
        self.catalog = catalog
        self.solution = solution
        self.iteration = iteration



#!/usr/bin/env python

import os
import sys
from collections import namedtuple

UserFolders = namedtuple("UserFolders", ("home", "desktop", "startmenu"))

uname = sys.platform
if os.name == "nt":
    uname = "win"
if uname == "linux2":
    uname = "linux"

from .linux import (scut_ext, ico_ext, make_shortcut,
                    get_folders, get_homedir, get_desktop)

if uname.startswith('win'):
    from .windows import (scut_ext, ico_ext, make_shortcut,
                          get_folders, get_homedir, get_desktop)

elif uname.startswith('darwin'):
    from .darwin import (scut_ext, ico_ext, make_shortcut,
                         get_folders, get_homedir, get_desktop)

from .shortcut import shortcut, Shortcut, fix_filename


def get_cwd():
    """get current working directory
    Note: os.getcwd() can fail with permission error.

    when that happens, this changes to the users `HOME` directory
    and returns that directory so that it always returns an existing
    and readable directory.
    """
    try:
        return os.getcwd()
    except:
        home = get_homedir()
        os.chdir(home)
        return home

try:
    import wx
    HAS_WX = True
    from .wxgui import ShortcutFrame
except ImportError:
    HAS_WX = False



import sys
from time import sleep
from unittest.mock import create_autospec

from PyQt5.QtCore import pyqtSignal, QThread
from PyQt5.QtWidgets import QDialog, QApplication, QVBoxLayout, QLabel, QProgressBar, \
    QPushButton
from album.api import Album

from album.gui.dialog.checklist_dialog import CheckListDialog
from album.gui.dialog.solution_action_process_thread import SolutionActionProcessThread
from album.gui.dialog.update_solutions_dialog import UpdateSolutionDialog
from album.gui.solution_util import generate_remove_solutions_actions


class UninstallSolutionsFromCatalogDialog(CheckListDialog):
    upgrade_running = pyqtSignal(bool)
    upgradeRunning = pyqtSignal(bool)
    upgradeDone = pyqtSignal()
    availableUpdatesChanged = pyqtSignal(bool)

    def __init__(self, options, album_instance):
        super().__init__(options, "Uninstall solutions from catalogs", "From which catalogs do you want to uninstall all solutions?", "Uninstall solution(s)")
        self.album_instance = album_instance
        self.upgrade_running.connect(lambda enabled: self.set_finished() if not enabled else None)
        self.actionSubmitted.connect(self.upgrade)

    def create_progress_view(self) -> QDialog:
        view = QDialog(self)
        view.setWindowTitle("Uninstalling solutions..")
        layout = QVBoxLayout()
        self.current_catalog_label = QLabel()
        self.close_button = QPushButton("Close")
        self.close_button.setVisible(False)
        self.close_button.clicked.connect(view.close)
        self.close_button.clicked.connect(self.close)
        layout.addWidget(self.current_catalog_label)
        self.progressbar = QProgressBar()
        self.progressbar.setRange(0, 0)
        layout.addWidget(self.progressbar)
        layout.addWidget(self.close_button)
        view.setLayout(layout)
        return view

    def upgrade(self, catalog_names):
        actions = generate_remove_solutions_actions(self.album_instance, catalog_names)
        self.solution_actions = actions
        self.current_processed_item = 0
        self.upgrade_running.emit(True)
        dialog = self.create_progress_view()
        self.upgrade_next_item()
        dialog.exec_()

    def upgrade_next_item(self):
        if self.current_processed_item >= len(self.solution_actions):
            self.upgrade_running.emit(False)
            return
        solution_action = self.solution_actions[self.current_processed_item]
        self.current_catalog_label.setText(UpdateSolutionDialog.get_action_text(solution_action))
        self.thread = SolutionActionProcessThread(solution_action, self.album_instance)
        self.thread.processFinished.connect(self.on_update_finished)
        self.thread.processFailed.connect(lambda x: self.upgradeDone.emit())
        self.thread.start()

    def on_update_finished(self, item):
        self.current_processed_item += 1
        self.upgrade_next_item()

    def set_finished(self):
        self.progressbar.setVisible(False)
        self.close_button.setVisible(True)
        self.current_catalog_label.setText("Uninstalling solutions done.")
        self.upgradeDone.emit()


class UpdateProcessThread(QThread):
    processFinished = pyqtSignal(str)

    def __init__(self, item, album_instance: Album):
        super().__init__()
        self.item = item
        self.album_instance = album_instance

    def run(self):
        self.album_instance.upgrade(self.item)
        self.processFinished.emit(self.item)


if __name__ == "__main__":
    app = QApplication(sys.argv)
    options = ["Option " + str(i) for i in range(1, 51)]
    album = create_autospec(Album)
    album.upgrade = lambda x: sleep(1)
    dialog = UninstallSolutionsFromCatalogDialog(options, album)
    if dialog.exec_() == QDialog.Accepted:
        all_checked, checked_items = dialog.getCheckedItems()
        print("All choices activated: ", all_checked)
        print("Activated choices: ", checked_items)
    sys.exit(app.exec_())


from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import QHBoxLayout, QWidget, QVBoxLayout, QLabel, QProgressBar, QTextEdit

from album.gui.widgets.util import create_btn, get_monospace_font


class SolutionStatus(QWidget):
    continue_signal = pyqtSignal()
    cancel_signal = pyqtSignal()

    def __init__(self, parent):
        super().__init__(parent)
        self.layout = QVBoxLayout(self)
        self.scroll_widget = QWidget()
        self.scroll_layout = QVBoxLayout(self.scroll_widget)
        self.run_output = QTextEdit()
        self.run_output.setReadOnly(True)
        self.run_output.setWordWrapMode(1)  # Word wrap
        self.run_output.setFont(get_monospace_font())
        self.layout.addWidget(self.run_output, 1)
        self.layout.addWidget(self._create_status_box())

    def _create_status_box(self):
        res = QWidget(self)
        actions_layout = QHBoxLayout(res)
        actions_layout.setContentsMargins(0, 0, 0, 0)
        self.status_label = QLabel(self._get_progress_text())
        self.status_widget = QProgressBar()
        self.status_widget.setRange(0, 0)
        # self.cancel_btn = create_btn(self, self.cancel_installation, "Cancel", "Esc")
        # self.cancel_btn.setVisible(True)
        self.continue_btn = create_btn(self, self.continue_signal, self._get_continue_text(), "Enter")
        self.continue_btn.setVisible(False)
        actions_layout.addWidget(self.status_label)
        actions_layout.addWidget(self.status_widget, 1)
        actions_layout.addStretch()
        # actions_layout.addWidget(self.cancel_btn)
        actions_layout.addWidget(self.continue_btn)
        return res

    def update_solution_log(self, record):
        if hasattr(record, "msg"):
            self.run_output.append(record.msg)
        else:
            self.run_output.append(record["msg"])
        # self.scroll.verticalScrollBar().setValue(self.scroll.verticalScrollBar().maximum())
        self.set_active()

    def set_solution_unfinished(self):
        self.status_widget.setVisible(True)
        self.status_label.setText(self._get_progress_text())
        self.continue_btn.setVisible(False)

    def set_solution_finished(self):
        self.status_widget.setVisible(False)
        self.status_label.setText(self._get_finished_text())
        self.continue_btn.setVisible(True)
        # self.cancel_btn.setVisible(False)

    def set_solution_failed(self):
        self.status_widget.setVisible(False)
        self.status_label.setText(self._get_failed_text())
        # self.cancel_btn.setVisible(False)

    def set_active(self):
        self.continue_btn.setDefault(True)
        self.continue_btn.setAutoDefault(True)

    def set_not_active(self):
        self.continue_btn.setDefault(False)
        self.continue_btn.setAutoDefault(False)

    def _get_continue_text(self):
        return "Continue"

    def _get_progress_text(self):
        return "Uninstalling..."

    def _get_failed_text(self):
        return "Failed."

    def _get_finished_text(self):
        return "Finished."



import sys
from copy import deepcopy
from unittest.mock import create_autospec

from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import QDialog, QApplication, QVBoxLayout, QLabel, QProgressBar, \
    QPushButton
from album.api import Album

from album.gui.dialog.checklist_table_dialog import CheckListTableDialog
from album.gui.dialog.solution_action_process_thread import SolutionActionProcessThread
from album.gui.dialog.update_action_item import UpdateAction
from album.gui.solution_util import group_solutions_by_version, full_coordinates, generate_update_actions


class SolutionsActionDialog(CheckListTableDialog):
    upgrade_running = pyqtSignal(bool)
    upgradeRunning = pyqtSignal(bool)
    upgradeDone = pyqtSignal()
    availableUpdatesChanged = pyqtSignal(bool)

    def __init__(self, album_instance, solution_actions, title, question, submit_text):
        super().__init__(solution_actions, title, question, submit_text)
        self.album_instance: Album = album_instance
        self.resize(600, 400)
        self.upgrade_running.connect(lambda enabled: self.set_finished() if not enabled else None)
        self.actionSubmitted.connect(self.upgrade)

    @staticmethod
    def _get_solution_options(index):
        actions = []
        catalogs = []
        for catalog in index["catalogs"]:
            catalog = deepcopy(catalog)
            catalog["solutions"] = group_solutions_by_version(catalog)
            catalogs.append(catalog)
        generate_update_actions(catalogs, actions)
        return actions

    def create_progress_view(self) -> QDialog:
        view = QDialog(self)
        view.setWindowTitle("Running solution actions..")
        layout = QVBoxLayout()
        self.current_catalog_label = QLabel()
        self.close_button = QPushButton("Close")
        self.close_button.setVisible(False)
        self.close_button.clicked.connect(view.close)
        self.close_button.clicked.connect(self.close)
        layout.addWidget(self.current_catalog_label)
        self.progressbar = QProgressBar()
        self.progressbar.setRange(0, 0)
        layout.addWidget(self.progressbar)
        layout.addWidget(self.close_button)
        view.setLayout(layout)
        return view

    def upgrade(self, update_items):
        self.solutions_with_actions = update_items
        self.current_processed_item = 0
        self.upgrade_running.emit(True)
        dialog = self.create_progress_view()
        self.upgrade_next_item()
        dialog.exec_()

    def upgrade_next_item(self):
        if self.current_processed_item >= len(self.solutions_with_actions):
            self.upgrade_running.emit(False)
            return
        solution_action = self.solutions_with_actions[self.current_processed_item]
        self.current_catalog_label.setText(self._get_action_text(solution_action))
        self.thread = SolutionActionProcessThread(solution_action, self.album_instance)
        self.thread.processFinished.connect(self.on_update_finished)
        self.thread.processFailed.connect(self.on_update_finished)
        self.thread.start()

    @staticmethod
    def _get_action_text(solution_action):
        action_str = ""
        if solution_action.update_action == UpdateAction.INSTALL:
            action_str = "Installing"
        if solution_action.update_action == UpdateAction.UNINSTALL:
            action_str = "Uninstalling"
        return "%s %s" % (action_str, full_coordinates(solution_action.solution, solution_action.catalog))

    def on_update_finished(self, item):
        self.current_processed_item += 1
        self.upgrade_next_item()

    def set_finished(self):
        self.progressbar.setVisible(False)
        self.close_button.setVisible(True)
        self.current_catalog_label.setText("Actions completed.")
        self.upgradeDone.emit()


if __name__ == "__main__":
    app = QApplication(sys.argv)
    album = create_autospec(Album)
    # album.upgrade = lambda x: sleep(1)
    # album.get_index_as_dict.return_value = TestInteractiveGUI._mock_index_dict_updates()
    dialog = SolutionsActionDialog(album, [], "title", "question?", "submit")
    if dialog.exec_() == QDialog.Accepted:
        checked_items = dialog.getCheckedItems()
        print("Activated choices: ", checked_items)
    sys.exit(app.exec_())


from PyQt5.QtCore import Qt
from PyQt5.QtGui import QContextMenuEvent
from PyQt5.QtWidgets import QListView, QMenu

from album.gui.solution_util import full_coordinates, installed, get_uninstall_actions_for_version


class SolutionView(QListView):

    def __init__(self, collection_widget):
        super().__init__()
        self.collection_widget = collection_widget
        self.setModel(self.collection_widget.proxy)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.select_first_item()

    def select_first_item(self):
        index = self.collection_widget.proxy.index(0, 0)
        self.setCurrentIndex(index)
        self.scrollTo(index)

    def contextMenuEvent(self, e: QContextMenuEvent):
        if self.selectionModel().selection().indexes():
            source_index = self.collection_widget.proxy.mapToSource(self.selectedIndexes()[0])
            item = self.collection_widget.search_model.item(source_index.row(), 0)
            menu = QMenu(self)

            versions = item.action.versions

            newest_version = versions[len(versions) - 1]
            if not installed(newest_version):
                install_action = menu.addAction("Install newest version")
                install_action.triggered.connect(
                    lambda b: self.collection_widget.install_solution.emit(full_coordinates(newest_version, item.action.catalog)))

            uninstall_old_versions_actions = get_uninstall_actions_for_version(self.collection_widget.album_instance, versions, item.action.catalog, exclude_newest=True)
            if len(uninstall_old_versions_actions) > 0:
                uninstall_old_action = menu.addAction("Uninstall old versions")
                uninstall_old_action.triggered.connect(lambda b: self.collection_widget.uninstall_solutions.emit(uninstall_old_versions_actions))

            uninstall_all_versions_actions = get_uninstall_actions_for_version(self.collection_widget.album_instance, versions, item.action.catalog, exclude_newest=False)
            if len(uninstall_all_versions_actions) > 0:
                uninstall_action = menu.addAction("Uninstall all versions")
                uninstall_action.triggered.connect(lambda b: self.collection_widget.uninstall_solutions.emit(uninstall_all_versions_actions))
            menu.exec_(e.globalPos())



from functools import total_ordering

from PyQt5.QtGui import QStandardItemModel, QStandardItem


class ActionsListModel(QStandardItemModel):

    @total_ordering
    class ActionItem(QStandardItem):
        recently_used = None
        text = None
        action = None
        solution_coordinates = None

        def _is_valid_operand(self, other):
            return hasattr(other, "text")

        def __eq__(self, other):
            if not self._is_valid_operand(other):
                return NotImplemented
            return self.text[2:].lower() == other.text[2:].lower()

        def __lt__(self, other):
            if not self._is_valid_operand(other):
                return NotImplemented
            return self.text[2:].lower() < other.text[2:].lower()


    def __init__(self, item_list=None, parent=None):
        self.item_list = item_list
        super().__init__(parent)

# TODO implement filter model https://www.walletfox.com/course/qsortfilterproxymodelexample.php to filter e.g. by catalog



import pkg_resources
from PyQt5.QtCore import pyqtSignal, Qt
from PyQt5.QtWidgets import QHBoxLayout, QScrollArea, QFrame, QWidget, QVBoxLayout, QLabel

from album.gui.widgets.util import create_btn


class PreRunSolutionWidget(QWidget):
    # events
    run_solution = pyqtSignal()
    cancel_solution = pyqtSignal()

    def __init__(self, parent):
        super().__init__(parent)
        self._load_plugins()
        # gui
        self.setStyleSheet(
            "PreRunSolutionWidget { border-top: 1px solid #bbbbbb; border-bottom: 1px solid #bbbbbb;}")
        self.setMinimumWidth(500)
        self.setContentsMargins(0, 0, 0, 0)
        self.layout = QVBoxLayout(self)
        scroll_widget = QWidget()
        scroll_widget.setContentsMargins(0, 0, 0, 0)
        self.scroll_layout = QVBoxLayout(scroll_widget)
        self.scroll_layout.setAlignment(Qt.AlignTop)
        self.scroll = QScrollArea()
        self.scroll.setFrameShape(QFrame.NoFrame)
        self.scroll.setWidget(scroll_widget)
        self.scroll.setWidgetResizable(True)
        self.scroll.setViewportMargins(0, 0, 0, 0)
        self.scroll.setContentsMargins(0, 0, 0, 0)
        self.required_warning = self._create_warning_msg()
        self.layout.addWidget(self.scroll)
        self.layout.addWidget(self.required_warning)
        self.layout.addWidget(self._create_actions_box())
        self.required_warning.setVisible(False)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self._widgets = {}
        self._required = []

    def _create_actions_box(self):
        res = QWidget()
        actions_layout = QHBoxLayout(res)
        actions_layout.setContentsMargins(5, 5, 5, 5)
        actions_layout.addWidget(create_btn(self, self.cancel_solution, "Cancel", "Esc"))
        self.run_btn = create_btn(self, self.run_solution, "Run", "Enter")
        self.run_btn.setDefault(True)
        self.run_btn.setAutoDefault(True)
        actions_layout.addWidget(self.run_btn)
        return res

    def _load_plugins(self):
        self._plugins = {}
        for entry_point in pkg_resources.iter_entry_points('album_gui_plugins'):
            # try:
            #     entry_point.load()
            #     self._plugins[entry_point.name] = entry_point.load()
            # except Exception as e:
            #     get_active_logger().error("Cannot load GUI plugin %s" % entry_point.name)
            #     get_active_logger().error(str(e))
            entry_point.load()
            self._plugins[entry_point.name] = entry_point.load()

    def add_arguments(self, arguments):
        self._widgets = {}
        for arg in arguments:
            self._add_argument(arg)
        max_height = 0
        for widget_name in self._widgets:
            widget = self._widgets[widget_name]
            max_height += widget.height()
        self.scroll.setMinimumHeight(min(max_height, 300))

    def _add_argument(self, argument):
        self.scroll.setFrameShape(QFrame.Box)
        arg_name = argument["name"]

        if ("required" not in argument or argument["required"]) and "default" not in argument:
            if arg_name not in self._required:
                self._required.append(arg_name)
        if "type" in argument:
            arg_type = argument["type"]
        else:
            arg_type = "string"
        if arg_type not in self._plugins:
            arg_type = "string"
        widget = self._plugins[arg_type](argument)

        self._widgets[arg_name] = widget
        self.scroll_layout.addWidget(widget)

    def get_values(self):
        res = {}
        for name in self._widgets:
            value = self._widgets[name].get_wrapped_value()
            if value is not None:
                res[name] = str(value)
        return res

    def check_required_fields(self) -> bool:
        values = self.get_values()
        missing_values = []
        for name in self._required:
            if name not in values or not values[name]:
                missing_values.append(name)
        if missing_values:
            self.required_warning.setVisible(True)
            self._highlight_missing_values(missing_values)
            self.required_warning.update()
            self.required_warning.parent().update()
            self.layout.update()
            return False
        else:
            self.required_warning.setVisible(False)
            self.layout.update()
        return True

    def _highlight_missing_values(self, values):
        for name in self._widgets:
            widget = self._widgets[name]
            if name in values:
                widget.setStyleSheet("QGroupBox{border: 1px solid red;}")
                widget.repaint()
            else:
                widget.setStyleSheet("")

    @staticmethod
    def _create_warning_msg():
        res = QLabel("Please provide the missing required arguments.")
        res.setWordWrap(True)
        res.setStyleSheet("color: red;")
        res.setMargin(5)
        return res

    def set_active(self):
        self.run_btn.setDefault(True)
        self.run_btn.setAutoDefault(True)

    def set_not_active(self):
        self.run_btn.setDefault(False)
        self.run_btn.setAutoDefault(False)



from PyQt5.QtGui import QFontInfo, QFont
from PyQt5.QtWidgets import QPushButton, QMessageBox


def display_error(text):
    msg = QMessageBox()
    msg.setIcon(QMessageBox.Critical)
    msg.setText("Error")
    msg.setInformativeText(text)
    msg.setWindowTitle("Error")
    msg.exec_()


def display_info(title, text):
    msg = QMessageBox()
    msg.setIcon(QMessageBox.Information)
    msg.setInformativeText(text)
    msg.setWindowTitle(title)
    msg.exec_()


def display_confirmation(text):
    msg = QMessageBox()
    msg.setIcon(QMessageBox.Question)
    msg.setText("Confirmation requested")
    res = msg.question(None, "Confirmation Requested", text, msg.Yes | msg.No)
    return res == msg.Yes


def create_btn(parent, emit_slot, label, shortcut):
    btn = QPushButton(parent)
    # box = QHBoxLayout()
    # box.setContentsMargins(4, 4, 4, 4)
    # box.addWidget(QLabel(label))
    # box.addStretch()
    # shortcut_label = QLabel("%s" % shortcut)
    # shortcut_label.setFont(get_monospace_font())
    # shortcut_label.setStyleSheet("color: rgba(0, 0, 0, 100);")
    # box.addWidget(shortcut_label)
    # btn.setLayout(box)
    btn.setText(label)
    btn.clicked.connect(emit_slot)
    return btn


def is_fixed_pitch(font):
    fi = QFontInfo(font)
    return fi.fixedPitch()


def get_monospace_font():
    font = QFont("monospace")
    if is_fixed_pitch(font):
        return font
    font.setStyleHint(QFont.Monospace)
    if is_fixed_pitch(font):
        return font
    font.setStyleHint(QFont.TypeWriter)
    if is_fixed_pitch(font):
        return font
    font.setFamily("courier")
    if is_fixed_pitch(font):
        return font
    return font



import packaging.version
from album.api import Album
from copy import deepcopy

from album.gui.dialog.update_action_item import UpdateActionItem, UpdateAction


def solution_version(solution):
    return solution["setup"]["version"]


def solution_coordinates(solution):
    coordinates = "%s:%s:%s" % (
        solution["setup"]["group"], solution["setup"]["name"], solution_version(solution))
    return coordinates


def installed(solution):
    return solution["internal"]["installed"]


def was_launched(solution):
    return bool(solution["internal"]["last_execution"])


def group_solutions_by_version(catalog):
    solutions = {}
    for solution in catalog["solutions"]:
        coordinates = "%s:%s:%s" % (
            catalog["name"], solution["setup"]["group"], solution["setup"]["name"])
        if coordinates in solutions:
            solutions[coordinates].append(solution)
        else:
            solutions[coordinates] = [solution]
    for solution_key in solutions:
        versioned_solutions = solutions[solution_key]

        solutions[solution_key] = sorted(versioned_solutions, key=lambda x: preprocess_version(x))
    return solutions


def preprocess_version(version):
    version = version["setup"]['version'].replace("-SNAPSHOT", ".dev0")
    try:
        return packaging.version.Version(version)
    except ValueError:
        return packaging.version.Version("0.0.0")



def full_coordinates(solution, catalog):
    coordinates = solution_coordinates(solution)
    return "%s:%s" % (catalog["name"], coordinates)


def is_parent(solution):
    return len(solution["internal"]["children"]) > 0


def child_installed(album: Album, solution):
    installed_children = [album._controller.collection_manager().get_collection_index().get_solution_by_collection_id(
        child["collection_id_child"]).internal()["installed"] for child in solution["internal"]["children"]]
    return any(installed_children)


def generate_update_actions(catalogs, actions, index=0):
    remaining_catalogs = []
    new_actions = []
    for catalog in catalogs:
        remaining_solutions = {}
        for solution_key in catalog["solutions"]:
            versions = catalog["solutions"][solution_key]
            solution_actions = []
            installed_versions_not_unlaunched_parents = False
            installed_old = []
            solution_used = False
            recent_version = versions[len(versions)-1]
            for idx, version in enumerate(reversed(versions)):
                if installed(version):
                    if was_launched(version):
                        solution_used = True
                    # check if a parent solution was ever launched and if not, if it is the parent of a solution to be uninstalled, if yes, mark for uninstallation too
                    if is_parent(version) and not was_launched(version):
                        can_be_uninstalled = True
                        for child in version["internal"]["children"]:
                            child_id = child["collection_id_child"]
                            child_solution = [action for action in actions if
                                              action.solution["internal"]["collection_id"] == child_id]
                            if len(child_solution) == 0 or child_solution[0].update_action is not UpdateAction.UNINSTALL:
                                # child solution is not marked to be uninstalled
                                if solution_key in remaining_solutions:
                                    remaining_solutions[solution_key].append(version)
                                else:
                                    remaining_solutions[solution_key] = [version]
                                can_be_uninstalled = False
                                break
                        if can_be_uninstalled:
                            solution_actions.append(
                                UpdateActionItem(UpdateAction.UNINSTALL, catalog, version, index))
                    else:
                        installed_versions_not_unlaunched_parents = True
                    if idx != 0 and not is_parent(version):
                        installed_old.append(version)
            if not installed(recent_version) and solution_used:
                solution_actions.append(UpdateActionItem(UpdateAction.INSTALL, catalog, versions[-1], index))
            if installed_versions_not_unlaunched_parents:
                for old_installation in installed_old:
                    solution_actions.append(UpdateActionItem(UpdateAction.UNINSTALL, catalog, old_installation, index))
            new_actions.extend(solution_actions)
        if len(remaining_solutions) > 0:
            catalog["solutions"] = remaining_solutions
            remaining_catalogs.append(catalog)
        actions.extend(new_actions)
        if len(remaining_catalogs) > 0 and len(new_actions) > 0:
            generate_update_actions(remaining_catalogs, actions, index + 1)


def generate_install_all_actions(album_instance, catalog_src):
    actions = []

    catalogs = []
    for catalog in album_instance.get_index_as_dict()["catalogs"]:
        if catalog["src"] == catalog_src:
            catalog = deepcopy(catalog)
            catalog["solutions"] = group_solutions_by_version(catalog)
            catalogs.append(catalog)

    for catalog in catalogs:
        for solution_key in catalog["solutions"]:
            versions = catalog["solutions"][solution_key]
            recent_version = versions[len(versions)-1]
            actions.append(UpdateActionItem(UpdateAction.INSTALL, catalog, recent_version))
    return actions


def generate_remove_solutions_actions(album_instance: Album, catalog_names):
    catalogs = []
    for catalog in album_instance.get_index_as_dict()["catalogs"]:
        if catalog["name"] in catalog_names:
            catalog = deepcopy(catalog)
            catalog["solutions"] = group_solutions_by_version(catalog)
            catalogs.append(catalog)

    uninstall_actions = []
    found_action = True
    while found_action:
        found_action = False
        for catalog in catalogs:
            for solution_key in catalog["solutions"]:
                versions = catalog["solutions"][solution_key]
                for idx, version in enumerate(reversed(versions)):
                    if not installed(version):
                        continue
                    if not is_parent(version) and version not in [v.solution for v in uninstall_actions]:
                        found_action = True
                        uninstall_actions.append(UpdateActionItem(UpdateAction.UNINSTALL, catalog, version))
                    if not child_installed(album_instance, version) and version not in [v.solution for v in uninstall_actions]:
                        found_action = True
                        uninstall_actions.append(UpdateActionItem(UpdateAction.UNINSTALL, catalog, version))
                    else:
                        for child in version["internal"]["children"]:
                            child_id = child["collection_id_child"]
                            child_will_be_uninstalled = any(
                                [to_be_uninstalled for to_be_uninstalled in uninstall_actions if
                                 to_be_uninstalled.solution["internal"]["collection_id"] == child_id])
                            if child_will_be_uninstalled and version not in [v.solution for v in uninstall_actions]:
                                found_action = True
                                uninstall_actions.append(UpdateActionItem(UpdateAction.UNINSTALL, catalog, version))
    return uninstall_actions


def get_newest_prefer_installed(versions):
    # check which is the newest installed version
    for version in reversed(versions):
        if installed(version):
            return version
    # no version installed, return the most recent one
    return versions[len(versions) - 1]


def get_uninstall_actions_for_version(album_instance, versions, catalog, exclude_newest=False):
    newest_installed = get_newest_prefer_installed(versions)
    if not installed(newest_installed):
        # no versions installed
        return []
    versions_to_be_uninstalled = []
    found_action = True
    while found_action:
        found_action = False
        for idx, version in enumerate(reversed(versions)):
            if not installed(version):
                continue
            if exclude_newest and version == newest_installed:
                continue
            if not child_installed(album_instance, version) and version not in versions_to_be_uninstalled:
                found_action = True
                versions_to_be_uninstalled.append(version)
            else:
                for child in version["internal"]["children"]:
                    child_id = child["collection_id_child"]
                    child_will_be_uninstalled = any([to_be_uninstalled for to_be_uninstalled in versions_to_be_uninstalled if
                                                     to_be_uninstalled["internal"]["collection_id"] == child_id])
                    if child_will_be_uninstalled and version not in versions_to_be_uninstalled:
                        found_action = True
                        versions_to_be_uninstalled.append(version)
    return [UpdateActionItem(UpdateAction.UNINSTALL, catalog, version) for version in versions_to_be_uninstalled]
